#include "pkgs.h"
#include "thread.h"
#include "tsocket.h"
#include "util.h"
#include "test-proxy.h"

/*

  Description of phase II:

  Client:
    1) Open 200 connections (not simultaneously)
    2) Write data
    3) Read it back
    3) Close connections


 */


static int num_threads;
static condition_t thread_done;
int error_flag = 0;
static int max_client_threads;

int client_check(int fd, check_data_p data)

{
  char my_buf[1024];
  int  len, cnt;
  int  bytes_left = sample_size - data->offset;

  if (!bytes_left)
	  return (0);
  talarm(5);
  if ((len = tread(fd, my_buf, sizeof(my_buf))) <= 0) {
    talarm(0);
    if (len < 0)
      prerror ("client_check: read:");
		
    return len;
  }

  talarm(0);

  if (data->offset == 0) {
    data->diff = (int)sample_buffer[0]  - (int)my_buf[0];
  }

  if (bytes_left < len) {
    fprintf (stderr, "Too many bytes read\n");
    return (-1);
  }
		
  for (cnt = 0; cnt < len; cnt++) {
    if ((char)(data->diff + (int)my_buf[cnt]) !=
	(char)sample_buffer[data->offset + cnt]) {
      fprintf (stderr, "%d %d\n", (data->diff + (int)my_buf[cnt]),
	       (int)sample_buffer[data->offset + cnt]);
      fprintf (stderr, "Data corruption: %d %d\n", data->offset + cnt,
	       data->diff);
      return (-1);
    }
  }

  data->offset += len;
  return (len);
}

static int num_conn = 0;
static int max_num_conn = 1;
int done = 0;
static int num_client_threads = 0;
static int accept_flag = 0;
static condition_t accept_conn;
static phase_info_t ginfo;

static void client_stub (void *arg)

{
	int fd = (int)arg;

	(*ginfo.client)(arg);

	tclose(fd);
	num_client_threads--;num_threads--;
	condition_broadcast(&thread_done);
}

static void client(void *arg)

{
  int   fd;

  num_conn = 0;
  num_client_threads = 0;

  while (num_conn < max_num_conn)
    {
      num_conn++;
      while (num_client_threads >= max_client_threads)
	      if (condition_wait(&thread_done)) {
		      prerror ("condition interrupted");
		      goto err_exit;
	      }

      talarm(10);

      fd = tsocket(AF_INET, SOCK_STREAM, 0);
      if (fd < 0) {
	prerror ("client: tsocket");
	exit (-1);
      }

      accept_flag = 0;
      if (tconnect (fd, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
	prerror ("connect");
	tclose (fd);
	goto err_exit;
      }

      while (!accept_flag) {
	      if (condition_wait(&accept_conn) < 0) {
		      prerror ("accept wait interrupted");
		      tclose (fd);
		      goto err_exit;
	      }
      }
      talarm(0);

      num_client_threads++; num_threads++;
      thread_create(client_stub, (void *)fd);
    }

  while (num_client_threads)
	  if (condition_wait(&thread_done))
		  goto err_exit;

  goto exit;

 err_exit:
  error_flag = 1;
 exit:
  done = 1;

  num_threads--;
  condition_broadcast(&thread_done);
#if 0
  printf ("Client done\n");
#endif
}


/*** SERVER SIDE OF TEST PROGRAM **/

int accept_launched = 0;
condition_t accept_cond;
int num_server_threads;

static void server_thread_stub(void *arg)

{
	int fd = (int) arg;
	
	(*ginfo.server)(arg);

	tclose (fd);
	num_server_threads--; num_threads--;
	condition_broadcast(&thread_done);
}

static void server_accept_thread(void *arg)

{
	int  accept_socket;

	num_server_threads = 0;

	accept_socket = tcpserv(SERVER_PORT);

	if (accept_socket < 0) {
	     prerror ("server: tcpserv");
	     return;
	}

	accept_launched = 1;
	condition_broadcast(&accept_cond);

	while (1) {
		int  newfd;

		while (num_server_threads >= max_client_threads) {
			if (condition_wait(&thread_done)) {
				prerror ("server: thread_done:");
				goto cleanup;
			}
		}

		newfd = taccept(accept_socket, NULL, NULL);

		if (newfd < 0) {
			if (errno != EINTR) {
				prerror ("server: accept");
				error_flag = 1;
			}

			goto cleanup;
		}

		accept_flag = 1;
		condition_signal(&accept_conn);

		num_server_threads++; num_threads++;
		thread_create(server_thread_stub, (void *)newfd);
	}
 cleanup:
	tclose(accept_socket);

	num_threads--;
	condition_broadcast(&thread_done);
}

void do_phase(phase_info_t *info, phase_ret_t *ret)

{
	int  flag;

	error_flag = 0;
	accept_launched =0;
	done =0;

	max_num_conn = info->max_conn;
	max_client_threads = info->max_sim;

	ginfo = *info;

	num_threads++;
	thread_create(server_accept_thread, NULL);

	while (!accept_launched)
		condition_wait(&accept_cond);

	num_threads++;
	
	thread_create(client, NULL);

	while (!error_flag && !done) {
		condition_wait(&thread_done);
	}

	/* Shut everything down */
	ret->succeeded = !error_flag;
	ret->num_conn = num_conn;

	flag = error_set(0);

	thread_broadcastalert();
	while (num_threads)
		condition_wait(&thread_done);

	error_set(flag);

	return;
}


