#include "../config.h"
#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <stdarg.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sys/socket.h>

#include <signal.h>
#include <sys/signal.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>

#include "thread.h"
#include "fdlim.h"
#include "util.h"
#include "timeout.h"

#ifdef __sparc__

#ifndef HAVE_SYS_FRAME_H
#include <sun4/frame.h>
#include <sun4/asm_linkage.h>
#else
#include <sys/frame.h>
#include <sys/stack.h>
#endif

#endif /* __sparc__ */


#define INRANGE(x)  (!(x) || ((x) >= all_thread && (x) <= &all_thread[MAX_THREAD]))


static void  thread_schedule(void);
extern void thread_switch(void *, void *);
static void my_fdclr(int fd, int write);


/*
 * Conceived by M. Frans Kaashoek (kaashoek@mit.edu)
 *
 * Enhanced by C. Sapuntzakis to support all sorts of fun stuff
 *
 */



/*
 *		Global thread variables
 */

struct thread {
  /* Note: Stack pointer must be first */
  double  dummy;
  long	  sp;		/* current stack pointer */
  long	  pc;		/* program counter */
  char    *stack;
#if 0
  char    stack[STACK_SIZE];
#endif
  void	  (*func)(void *);      /* thread entry point */
  void    *arg;
  int	  state;		/* running, runnable, waiting */
  DEQUE_ENTRY(thread) link;
  threadq_p   queue;
  int         alerted;
  int         thread_errno;
};


threadq_t runnables;

/* For sharing with thread_switch command */
thread_p  current_thread;

static thread_t all_thread[MAX_THREAD];

int max_fd;

/* The wait queues */
static condition_t *rthread, *wthread;

static fd_set rfds, wfds;                   /* Bitmap of cb's in use */
static int numrfds, numwfds;
static struct timeval selwait;


#define threadq_head(threadq)     ((threadq)->first)
#define threadq_tail(threadq)     ((threadq)->last)

void check_invariant(void)

{
	{
	  thread_p t;
	  
	  /* Look for a free thread. */
	  for (t = all_thread; t < all_thread + MAX_THREAD; t++) {
	    if (t->state == FREE) {
	      continue;
	    }
	    assert (INRANGE(t->link.next) &&
		    INRANGE(t->link.prev));
	    
	  }
	}
}

void threadq_remove(thread_p thread)

{
	if (!thread->queue)
		return;

	DEQUE_REMOVE(thread->queue, thread, link);
	thread->queue = NULL;
}

void threadq_insert_head(threadq_p queue, thread_p thread)

{
	assert (!thread->queue);

	DEQUE_INSERT_HEAD(queue, thread, link);
	thread->queue = queue;
}

void threadq_insert_tail(threadq_p queue, thread_p thread)

{
	assert (!thread->queue);

	DEQUE_INSERT_TAIL(queue, thread, link);
	thread->queue = queue;
}

void
condition_init(condition_p var)

{
	memset (var, 0, sizeof (*var));
}

int
condition_wait(condition_p var)

{
	threadq_remove(current_thread);
	threadq_insert_tail(var, current_thread);

	current_thread->state = WAITING;
	thread_schedule();

	if (current_thread->alerted) {
		current_thread->alerted = 0;
		return (-1);
	} 

	return (0);
}

void
condition_signal(condition_p var)

{
	thread_p thread = threadq_head(var);

	if (thread) {
		threadq_remove(thread);
		thread->state = RUNNABLE;
		threadq_insert_tail(&runnables, thread);
	}
}


void 
condition_broadcast_internal(condition_p var, void (*func)(thread_p, void *),
			     void *arg)

{
	thread_p thread;

	while ((thread = threadq_head(var))) {
		threadq_remove(thread);
		thread->state = RUNNABLE;
		threadq_insert_tail(&runnables, thread);

		if (func)
			(*func)(thread, arg);

	}	
}

void
condition_broadcast(condition_p var)

{
	condition_broadcast_internal(var, NULL, NULL);
}

void
thread_alert(thread_p thread)

{
	thread->alerted = 1;
	/* FIXME: What should really happen? */
	thread->thread_errno = EINTR;

	threadq_remove(thread);
	thread->state = RUNNABLE;
	threadq_insert_tail(&runnables, thread);
}

void
thread_broadcastalert(void)

{
	int cnt;

	for (cnt = 0; cnt < MAX_THREAD; cnt++) {
		if (all_thread[cnt].state != FREE) {
			thread_alert(&all_thread[cnt]);
		}
	}
}


void
thread_alarm(thread_p thread, int sec)

{
	remove_timeout(thread);
	if (sec)
		add_timeout(sec, thread);
}


void
shutdown_callback(thread_p thread, void *arg)

{
	thread->alerted = 1;
	thread->thread_errno = EIO;
}

#define OPEN_READ  1
#define OPEN_WRITE 2
char open_fd[1024];

void
fd_check(int s, int write)

{
  char tst = open_fd[s];
  assert ((write ? (tst & OPEN_WRITE) : (tst & OPEN_READ)));
}

void
fd_open(int s)

{
  open_fd[s] = OPEN_READ | OPEN_WRITE;
}

void
fd_shutdown(int  s, int how)

{
	if (!is_validfd(s))
		return;

	if (how == 0 || how == 2) {
		my_fdclr(s, 0);

		open_fd[s] &= ~OPEN_READ;

		condition_broadcast_internal(&rthread[s], shutdown_callback,
					     NULL);
	}

	if (how == 1 || how == 2) {
		my_fdclr(s, 1);

		open_fd[s] &= ~OPEN_WRITE;

		condition_broadcast_internal(&wthread[s], shutdown_callback,
					     NULL);
	}
}


static void
my_fdset(int fd, int write)

{
	if (write) {
	         
		if (!FD_ISSET(fd, &wfds))
			numwfds++;
		FD_SET(fd, &wfds);
	} else {
		if (!FD_ISSET(fd, &rfds))
			numrfds++;
		FD_SET(fd, &rfds);
	}
}

static void
my_fdclr(int fd, int write)

{
	if (!write) {
		if (!FD_ISSET(fd, &rfds)) 
			return;

		FD_CLR(fd, &rfds);
		numrfds--;
	} else {
		if (!FD_ISSET(fd, &wfds)) 
			return;

		FD_CLR(fd, &wfds);
		numwfds--;
	}
}
		
#if 0		
static int
my_fdisset(int fd, int write)

{
	return (write ? FD_ISSET(fd, &wfds) : FD_ISSET(fd, &rfds));
}
#endif
int
is_validfd(int fd)

{
	return ((fd >= 0) && (fd <= max_fd));
}			

int
fd_sleep(int fd, int write)

{
	condition_p slot = (write ? &wthread[fd] : &rthread[fd]);
	int  signal;

	check_invariant();
	assert ((fd >=0) && (fd <= max_fd));

	fd_check(fd, write);
	my_fdset(fd, write);

	signal = condition_wait(slot);

	return (signal);
}

fd_set trfds, twfds;

int check_io (int do_wait)

{
  thread_p  thread;
  int i, n = 0;


try_again:

  selwait.tv_sec = 0;
  selwait.tv_usec = 0;

  if (do_wait) {
	  int  wait;
	  if ((wait = next_timeout()) == -1)
		  selwait.tv_sec = 600;
	  else
		  selwait.tv_sec = wait;
	  selwait.tv_usec = 0;
  }

  /* Call select.  Since the fd_sets are both input and output
   * arguments, we must copy rfds and wfds. */
  trfds = rfds;
  twfds = wfds;

  n = select (max_fd, &trfds, &twfds, NULL, &selwait);
  if (n < 0) {
	  if (errno == EINTR)
		  goto try_again;
    fatal ("select: %s\n", strerror (errno));
  }

  if ((thread = get_timeout())) {
	  do {
		  thread_alert(thread);
		  thread->thread_errno = EALARM;

	  } while ((thread = get_timeout()));
  } else if (n == 0 && do_wait) {
	  if (numrfds == 0 && numwfds == 0)
		  return (0);

	  goto try_again;
  }

  /* Loop through and make callbacks for all ready file descriptors */
  for (i = 0; n && i < max_fd; i++) {
    if (FD_ISSET (i, &trfds)) {
      n--;
      /* Because any one of the callbacks we make might in turn call
       * cb_free on a higher numbered file descriptor, we want to make
       * sure each callback is wanted before we make it.  Hence check
       * rfds. */
      if (FD_ISSET (i, &rfds)) {
	      my_fdclr(i, 0);
	      condition_broadcast(&rthread[i]);
      }
    }
    if (FD_ISSET (i, &twfds)) {
      n--;
      if (FD_ISSET (i, &wfds)) {
	      my_fdclr(i, 1);
	      condition_broadcast(&wthread[i]);
      }
    }
  }

  return (1);

}



void 
thread_init(void)
{
    sigset_t sm;
    int size;

    sigemptyset(&sm);
    sigaddset(&sm, SIGPIPE);  
    sigprocmask (SIG_BLOCK, &sm, NULL);

    FD_ZERO(&rfds);
    FD_ZERO(&wfds);

    FD_ZERO(&trfds);
    FD_ZERO(&twfds);

    /* We'll probably need more than 64 file descriptors */
    if (fdlim_set(FD_SETSIZE) < 0)
	    printf ("fdlim failed\n");

    max_fd = fdlim_get(0);

    if (max_fd > FD_SETSIZE)
      max_fd = FD_SETSIZE;

    size = sizeof(condition_t) * max_fd;
    rthread = xmalloc(size);
    memset (rthread, 0, size);
    wthread = xmalloc(size);
    memset (wthread, 0, size);

    current_thread = &all_thread[0];
    current_thread->state = RUNNING;
    threadq_insert_head(&runnables, current_thread);
}


static void 
thread_schedule(void)
{
    int  res;
    thread_p  next_thread;

    current_thread->thread_errno = errno;

    check_io (0);

    /* Find another runnable thread. (round-robin) */
try_again:

    next_thread = threadq_head(&runnables);
    if (next_thread) {
	    threadq_remove(next_thread);
	    threadq_insert_tail(&runnables, next_thread);
    } else {
	    res = check_io(1);

	    if (res == 0) {
		    printf("thread_schedule: no runnable threads; deadlock\n");
	            exit(1);
	    }
	    goto try_again;
    }

    if (current_thread != next_thread) { 	/* switch threads?  */
        thread_p prev_thread = current_thread;
        current_thread = next_thread;

	current_thread->state = RUNNING;
	check_invariant();
	thread_switch(&prev_thread->sp, &current_thread->sp);
    }

    check_invariant();
    /* Restore errno global */
    errno = current_thread->thread_errno;
}

static void 
thread_terminate(void)
{
    current_thread->state = FREE;
    threadq_remove(current_thread);
    remove_timeout(current_thread);
    thread_schedule();
}


static void 
thread_body()
{
    thread_p t;

    check_invariant();

    t = current_thread;
    (*(t->func))(t->arg);
    thread_terminate();
}

void 
thread_print(thread_p t)
{
#if 0
    printf("current_thread: %d (0x%p)\n", current_thread - all_thread,
	   current_thread);
    printf("next_thread: %d (0x%p)\n", next_thread - all_thread, next_thread);
#endif
    printf("%d: sp 0x%lX pc 0x%lX state 0x%x \n", t - all_thread, 
	       t->sp, t->pc, t->state);
}


static thread_p
thread_find(void)
{
    thread_p t;

    /* Look for a free thread. */
    for (t = all_thread; t < all_thread + MAX_THREAD; t++) {
	if (t->state == FREE) {
		break;
	}
    }

    return t;
}

    
thread_p
thread_create(void (*func)(void *), void *arg)
{
    thread_p t;
#if __i386__
    int  *i;
#endif
#if __sparc__
    struct frame *sp;
#endif

    /* Found a free thread; init it. */

    t = thread_find();
    assert(t < all_thread + MAX_THREAD);

    if (!t->stack) {
      int fd;

      if ((fd = open("/dev/zero", O_RDWR)) < 0) {
	perror ("/dev/zero:");
	exit(1);
      }

      t->stack = mmap(NULL, 8192 + STACK_SIZE, PROT_READ | PROT_WRITE,
		      MAP_PRIVATE, fd, 0);

      close (fd);

      assert (t->stack && (int) t->stack != -1);

      mprotect(t->stack, 4096, PROT_NONE);
      t->stack += 4096;
      mprotect(t->stack + STACK_SIZE, 4096, PROT_NONE);
    }

#if __sparc__
    sp = (struct frame *) (t->stack + SA(STACK_SIZE - BODY_OFFSET 
						   + WINDOWSIZE));
    sp->fr_savfp = (struct frame *) (t->stack + SA(STACK_SIZE -
						   BODY_OFFSET));
    t->sp = (long) sp;
    t->pc = (long) thread_body - RETURN_OFFSET;
#endif
#if __i386__

    t->sp = (long) (t->stack + STACK_SIZE - (RETURN_OFFSET + 1) * sizeof(int));
    i = (int *) (t->sp);
    t->pc = (long) thread_body;
    i[RETURN_OFFSET] = t->pc;
#endif
    t->state = RUNNABLE;
    t->func = func;
    t->arg = arg;
    t->thread_errno =0;

    threadq_insert_tail(&runnables, t);
    return (t);
}

	
