#include <unistd.h>
#include <sys/times.h>
#include <time.h>
#include "timeout.h"
#include "util.h"

typedef struct tmo tmo_t, *tmo_p;

struct tmo {
	clock_t time;
	void *info;
	tmo_p next;
};

static tmo_p time_queue = NULL;

void add_timeout(int sec, void *info)

{
	clock_t evtime = time(NULL) + sec;
	tmo_p *prev = &time_queue;
	tmo_p next = time_queue;
	tmo_p new = xmalloc(sizeof(tmo_t));

	new->time = evtime;
	new->info = info;

	while (next) {
		if (next->time > evtime)
			break;

		prev = &next->next;
		next = next->next;
	}

	new->next = *prev;
	*prev = new;	       
}

void remove_timeout(void *info)

{
	tmo_p *prev = &time_queue;
	tmo_p next = time_queue;
	
	while (next) {
		if (next->info == info)
			break;

		prev = &next->next;
		next = next->next;
	}
	
	if (next) {
		*prev = next->next;
		xfree (next);
	}
}

int next_timeout()

{
	if (time_queue) {
		clock_t now = time(NULL);
		clock_t head_time = time_queue->time;

		if (head_time < now)
			return (0);
		else
			return (head_time - now);
	} else 
		return (-1);

}

void *get_timeout()

{
	if (time_queue) {
		clock_t now = time(NULL);
		void *info = time_queue->info;

		if (time_queue->time <= now) {
			tmo_p tmp = time_queue;

			time_queue = time_queue->next;

			xfree(tmp);
			return (info);
		}
	}

	return (NULL);
}
