#include "pkgs.h"
#include "thread.h"
#include "tsocket.h"
#include "util.h"
#include "test-proxy.h"

/*

  Description of phase I:

  Client:
    1) Open 200 connections (not necessarily simultaneously)
    2) Write data
    3) Read it back, check its integrity
    3) Close connections (there should be no timeout)


 */


static void client_thread(void *arg)

{
	int cnt, ret, fd;
	check_data_t check_data;
	int written;

	memset (&check_data, 0, sizeof(check_data));

	fd = (int) arg;

#define INCR 1024
	for (cnt = 0; cnt < sample_size; cnt += INCR) {
	        talarm(5);

		if ((written = twrite (fd, sample_buffer + cnt, INCR)) < 0) {
			prerror ("write");
			abort();
			goto err_exit;
		}
		assert(written == INCR);
		talarm(0);

		if (client_check(fd, &check_data) < 0) {
			goto err_exit;
		}
	}

	while ((ret = client_check(fd, &check_data)) > 0);

	if (ret == 0 && check_data.offset != cnt) {
		printf ("Offset Error: %d %d\n", check_data.offset, cnt);
		goto err_exit;
	} 

	if (ret < 0) {
		goto err_exit;
	}

	goto exit;

 err_exit:
	error_flag = 1;
 exit:
	talarm(0);
	return;
}

static int thread_id = 0;

static void server_thread(void *arg)

{
	int fd = (int)arg;
	char bytes[256];
	int len;
	int my_thread_id = thread_id++;
	int written;
	/* Do whatever we need to test it */
	while ((len = tread(fd, bytes, sizeof(bytes))) > 0) {
		int  idx;
		talarm (0);

		for (idx = 0; idx < len; idx++) {
		        /* modify the data slightly */
			bytes[idx] += my_thread_id;
		}

		written = twrite(fd,bytes,len);
	        if((written<0) && (errno!=EAGAIN)) {
			len = written;
			goto srv_err_exit;
	        }
		assert(written == len);

	}

srv_err_exit:
	if (len < 0) {
	  prerror ("read/write");
	  error_flag = 1;
	}

}

void do_phase1()

{
	phase_info_t  info = { client_thread, server_thread, 1, 1 };
	phase_ret_t ret; 
	int phase;

	printf ("Phase 1: (client write,server echo)checking the integrity of returned data\n");

	for (phase = 0; phase < 3; phase++) {
		switch (phase) {
		case 0:
			info.max_conn =200;
			info.max_sim = 1;
			break;
		case 1:
			info.max_conn =200;
			info.max_sim = 20;
			break;
		case 2:
			info.max_conn =200;
			info.max_sim = 70; 
			break;
		}

		printf ("%d simultaneous connections (with read/write), %d total...",
			info.max_sim, info.max_conn);

		do_phase(&info, &ret);

		if (!ret.succeeded) {
			printf ("Failed\n");
			kill_proxy();
			restart_proxy(proxy_name);
		} else 
			printf ("Succeeded\n");
      	}

}


