#include "pkgs.h"
#include "util.h"
#include "tsocket.h"
#include "thread.h"



/* Extremely primitive mechanism for suppressing extraneous error messages */
static int error_print = 1;

#ifndef HAVE_STRERROR
extern int sys_nerr;
extern char *sys_errlist[];

char *strerror(int errnum)

{
	if (errnum < sys_nerr)
		return (sys_errlist[errnum]);
	else
		return ("Unknown Error");
}

#endif

int error_set(int flag)

{
	int  s = error_print;
	error_print = flag;
	return (s);
}

void
prerror (const char *msg)

{
	if (error_print)
		if (errno == EALARM) {
			printf ("%s: Timer expired\n", msg);
		} else {
			printf ("%s: %s\n", msg, strerror(errno));
		}
}

/* Create a TCP socket, bind it to a particular port, and call listen
 * for connections on it.  These are the three steps necessary before
 * clients can connect to a server.  */
int
tcpserv (int port)
{
  int s, n;
  struct sockaddr_in saddr;

  /* The address of this server */
  bzero (&saddr, sizeof (saddr));
  saddr.sin_family = AF_INET;
  saddr.sin_port = htons (port);
  /* We are interested in listening on any and all IP addresses this
   * machine has, so use the magic IP address INADDR_ANY. */
  saddr.sin_addr.s_addr = htonl (INADDR_ANY);

  s = tsocket (AF_INET, SOCK_STREAM, 0);
  if (s < 0) {
    prerror ("socket");
    return -1;
  }

  /* Allow the program to run again even if there are old connections
   * in TIME_WAIT.  This is the magic you need to do to avoid seeing
   * "Address already in use" errors when you are killing and
   * restarting the daemon frequently.  */
  n = 1;
  if (setsockopt (s, SOL_SOCKET, SO_REUSEADDR, (char *)&n, sizeof (n)) < 0) {
    prerror ("SO_REUSEADDR");
    tclose (s);
    return -1;
  }

  if (tbind (s, (struct sockaddr *) &saddr, sizeof (saddr)) < 0) {
    fprintf (stderr, "TCP port %d: %s\n", port, strerror (errno));
    tclose (s);
    return -1;
  }
  if (tlisten (s, 5) < 0) {
    prerror ("listen");
    tclose (s);
    return -1;
  }

  return s;
}
